@@ -1,6 +1,8 @@
 # pylint: disable=no-self-use
 import json
 import time
+import warnings
+from concurrent.futures.thread import ThreadPoolExecutor
 from datetime import datetime
 from urllib.error import URLError
 
@@ -9,7 +11,7 @@ import requests
 
 from sdc_dp_helpers.api_utilities.date_managers import date_handler
 from sdc_dp_helpers.api_utilities.file_managers import load_file
-from sdc_dp_helpers.api_utilities.retry_managers import retry_handler
+from sdc_dp_helpers.api_utilities.retry_managers import retry_handler, request_handler
 
 
 class CustomOneSignalReader:
@@ -24,8 +26,11 @@ class CustomOneSignalReader:
         }
 
         self._request_session = requests.Session()
+        self.offset, self.data_set = 0, []
 
-    @retry_handler(exceptions=KeyError, total_tries=5, should_raise=True)
+    # ConnectionError should be intentional and are handled with a retry
+    # anything else is an unexpected issue that should not be retried but raised
+    @retry_handler(exceptions=ConnectionError, total_tries=10, should_raise=True)
     def _csv_export_query(self):
         """
         Query handler for csv_export.
@@ -58,23 +63,23 @@ class CustomOneSignalReader:
         if errors:
             # The header is definitely valid, so we must assume the app_id is not
             if (
-                errors
-                == "app_id not found. You may be missing a Content-Type: application/json header."
+                    errors
+                    == "app_id not found. You may be missing a Content-Type: application/json header."
             ):
-                raise EnvironmentError(
+                raise ConnectionError(
                     "App Id may not be valid or removed from the account."
                 )
 
             # Seems like we are restricted to making one query at a time for a given user, so wait 5 minutes
             if (
-                errors
-                == "User already running another CSV export. Please wait until your other CSV exporter finishes and try again."
+                    errors
+                    == "User already running another CSV export. Please wait until your other CSV exporter finishes and try again."
             ):
                 time.sleep(300)  # the reader may try again after the 5 minute wait
-                raise EnvironmentError(errors)
+                raise ConnectionError(errors)
 
             # Raise all other unknown errors in a standard fashion
-            raise EnvironmentError(errors)
+            raise ConnectionError(errors)
         else:
             print("No errors raised, fetching url.")
 
@@ -99,31 +104,87 @@ class CustomOneSignalReader:
                     )
                     data_frame = data_frame[
                         (
-                            data_frame["tmp_date"]
-                            >= datetime.strptime(sd, _date_fmt).strftime(_date_fmt)
+                                data_frame["tmp_date"]
+                                >= datetime.strptime(sd, _date_fmt).strftime(_date_fmt)
                         )
                         & (
-                            data_frame["tmp_date"]
-                            <= datetime.strptime(ed, _date_fmt).strftime(_date_fmt)
+                                data_frame["tmp_date"]
+                                <= datetime.strptime(ed, _date_fmt).strftime(_date_fmt)
                         )
-                    ]
+                        ]
                     data_frame.drop("tmp_date", axis="columns", inplace=True)
                 break
             except URLError:
-                # max wait time is 15 minutes at 10 attempts before we need to contact Onesignal
+                # max wait time is 15 minutes at 10 attempts before
+                # we need to contact Onesignal.
+                # these attempts are NOT traditional retry attempts
+                # so are not handed by @retry_handler.
                 print(f"Waiting for file to generate, attempt {attempts}.")
                 time.sleep(90)
 
             attempts += 1
 
             if attempts >= 10:
-                raise URLError(
+                raise FileExistsError(
                     f"Csv file could not be generated, contact Onesignal support. "
                     f"Response {response}."
                 )
 
         return data_frame.to_json(orient="records")
 
+    def _get_view_notification_payloads(self):
+        """
+        Generate all payloads for the request methods.
+        """
+        print("GET: view_notifications.")
+        initial_response = self._request_session.get(
+            url="https://onesignal.com/api/v1/notifications",
+            headers=self._header,
+            json={"app_id": self._creds.get("app_id")},
+        )
+
+        if initial_response.status_code == 200:
+            print(f"Connection made: {initial_response.status_code} and {initial_response.reason}.")
+            payloads, offsets = [], initial_response.json().get("total_count")
+            print(f"Total offsets: {offsets}.")
+            if offsets > 0:
+                for offset in range(0, offsets):
+                    payloads.append({"app_id": self._creds.get("app_id"), "offset": offset})
+                return payloads
+            else:
+                return None
+
+        if initial_response.status_code >= 500:
+            raise ConnectionError(
+                f"Connection failed with {initial_response.status_code} and connection closed due to a timeout."
+            )
+
+        raise ConnectionError(
+            f"Connection failed with {initial_response.status_code} and {initial_response.reason}."
+        )
+
+    # Onesignal suggest a 60 second wait for any errors above 500,
+    # these are internal timeouts that they are aware of.
+    @request_handler()
+    @retry_handler(exceptions=ConnectionError, total_tries=10, should_raise=True, initial_wait=60, backoff_factor=1)
+    def _view_notifications_get(self, payload):
+        """
+        Standard view notifications get request handler.
+        """
+        resp = self._request_session.get(
+            url="https://onesignal.com/api/v1/notifications",
+            headers=self._header,
+            json=payload,
+        )
+        print(f"Payload: {payload}. Status: {resp.status_code} - {resp.reason}.")
+        if resp.status_code == 200:
+            data_response = resp.json().get("notifications", None)
+            if data_response is not None:
+                return data_response
+            return None
+        else:
+            raise ConnectionError(f"Connection failed with {resp.status_code} and {resp.reason}.")
+
     def _view_notifications_query(self):
         """
         Query handler for view_notifications.
@@ -137,37 +198,40 @@ class CustomOneSignalReader:
 
         ⚠ Note that adding any date oriented payload drastically affects
         the output data. Even if the payload values are null.
+
+        With this method comes the option of running a ThreadPoolExecutor
+        that will query the payloads in batches of 10. Making the response
+        a lot quicker and reducing the change of timeouts and intermittent
+        connection issues from the server.
+        Python threads are really only useful for concurrent I/O operations.
+
+        WARNING: Setting too many threads will result in a request error
+                 of too many connection attempts.
         """
-        print("GET: view_notifications.")
-        initial_response = self._request_session.get(
-            url="https://onesignal.com/api/v1/notifications",
-            headers=self._header,
-            json={"app_id": self._creds.get("app_id")},
-        )
-
-        total_offset = initial_response.json().get("total_count")
-
-        data_set = []
-        for offset in range(0, total_offset):
-            print(f"At offset: {offset} of {total_offset}.")
-            response = self._request_session.get(
-                url="https://onesignal.com/api/v1/notifications",
-                headers=self._header,
-                json={"app_id": self._creds.get("app_id"), "offset": offset},
-            )
-
-            if response.status_code == 200:
-                # if no more data is present skip response
-                data = response.json().get("notifications", None)
-                if data is None:
-                    continue
-                data_set.append(data)
+        json_payloads = self._get_view_notification_payloads()
+        parallel = self._config.get("parallel", False)
+        print(f"Parallel connection: {parallel}.")
+        if json_payloads is not None:
+            if parallel:
+                # Run concurrent connections.
+                max_threads = self._config.get("max_threads", 15)
+                if max_threads >= 20:
+                    warnings.warn(
+                        "Too many threads. Onesignal Connection failed with 429 and Too Many Requests could be raised."
+                    )
+                with ThreadPoolExecutor(max_workers=max_threads) as pool:
+                    # append all responses to the dataset
+                    data_set = pool.map(self._view_notifications_get, json_payloads)
+                    for data in data_set:
+                        self.data_set.append(data)
             else:
-                raise requests.exceptions.ConnectionError(
-                    f"Connection failed with {response.status_code} and {response.reason}."
-                )
+                # Run a stock-standard request process.
+                for payload in json_payloads:
+                    self.data_set.append(
+                        self._view_notifications_get(payload)
+                    )
 
-        return json.dumps(data_set)
+        return json.dumps(self.data_set)
 
     def run_query(self):
         """
@@ -175,6 +239,7 @@ class CustomOneSignalReader:
         access data in Onesignal for view_notifications and csv_exports.
         """
         _endpoint = self._config.get("endpoint")
+        _parallel = self._config.get("parallel", False)
         if _endpoint == "csv_export":
             return self._csv_export_query()
         elif _endpoint == "view_notifications":
